import asyncio
import logging
from bisect import insort
from contextlib import AsyncExitStack, asynccontextmanager, contextmanager
from contextvars import ContextVar
from dataclasses import KW_ONLY, dataclass, field
from functools import wraps
from heapq import merge
from importlib.util import module_from_spec, spec_from_file_location
from inspect import (
    Parameter,
    isasyncgenfunction,
    isclass,
    iscoroutinefunction,
    isgeneratorfunction,
    signature,
)
from operator import attrgetter
from pathlib import Path
from types import UnionType
from typing import Annotated, Any, Callable, ClassVar, Union, get_args, get_origin

from fastbot.event import Context, Event
from fastbot.matcher import Matcher


@dataclass(slots=True)
class Plugin:
    @dataclass(slots=True)
    class Middleware:
        _: KW_ONLY

        priority: int = 0
        executor: Callable[..., Any]

    _: KW_ONLY

    state: ContextVar[bool] = ContextVar("state", default=True)

    init: Callable[..., Any] | None = None
    backgrounds: list[Callable[..., Any]] = field(default_factory=list)

    middlewares: list[Middleware] = field(default_factory=list)
    executors: list[Callable[..., Any]] = field(default_factory=list)

    async def run(self, *, event: Event) -> list[Any]:
        return await asyncio.gather(*(executor(event) for executor in self.executors))


@dataclass(slots=True)
class Dependency:
    _: KW_ONLY

    dependency: Callable[..., Any]

    @classmethod
    def provide(cls, dependency: Callable[..., Any]) -> Any:
        return cls(dependency=dependency)


class PluginManager:
    __slots__ = ()

    plugins: ClassVar[dict[str, Plugin]] = {}

    @classmethod
    def import_from(cls, path_to_import: str) -> None:
        def load(module_name: str, module_path: Path) -> None:
            module_name = module_name.removesuffix(".py")

            cls.plugins[module_name] = plugin = Plugin()

            try:
                spec = spec_from_file_location(module_name, module_path)
                module = module_from_spec(spec)  # type: ignore
                spec.loader.exec_module(module)  # type: ignore

                plugin.init = getattr(module, "init", None)

                logging.info(f"loaded plugin [{module_name}] from [{module_path}]")

            except Exception as e:
                logging.exception(e)

            finally:
                if not (
                    plugin.init
                    or plugin.backgrounds
                    or plugin.middlewares
                    or plugin.executors
                ):
                    del cls.plugins[module_name]

        if (path := Path(path_to_import)).is_dir():
            for file in path.rglob("*.py"):
                if file.is_file() and not file.name.startswith("_"):
                    load(".".join(file.relative_to(path.parent).parts), file)

        elif (
            path.is_file()
            and path.name.endswith(".py")
            and not path.name.startswith("_")
        ):
            load(".".join(path.parts).removesuffix(".py"), path)

    @classmethod
    async def run(cls, *, ctx: Context) -> None:
        try:
            for middleware in merge(
                middleware
                for plugin in cls.plugins.values()
                for middleware in plugin.middlewares
            ):
                _ = asyncio.Task(
                    middleware.executor(ctx),
                    loop=asyncio.get_running_loop(),
                    eager_start=True,
                )

                if not ctx:
                    logging.warning("context is empty, discarding")

                    return

            event = Event.from_ctx(ctx=ctx)

            await asyncio.gather(
                *(
                    plugin.run(event=event)
                    for plugin in cls.plugins.values()
                    if plugin.state.get()
                )
            )

        except Exception as e:
            logging.exception(e)


def background(func: Callable[..., Any]) -> Callable[..., Any]:
    PluginManager.plugins[func.__module__].backgrounds.append(func)

    return func


def middleware(*, priority: int = 0) -> Callable[..., Any]:
    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        insort(
            PluginManager.plugins[func.__module__].middlewares,
            Plugin.Middleware(priority=priority, executor=func),
            key=attrgetter("priority"),
        )

        return func

    return decorator


def on(matcher: Matcher | Callable[..., bool] | None = None) -> Callable[..., Any]:
    def annotation_event_type(annotation: Any) -> tuple[type[Event], ...]:
        if get_origin(annotation) in (Annotated, Union, UnionType):
            return tuple(
                arg
                for arg in get_args(annotation)
                if isclass(arg) and issubclass(arg, Event)
            )

        elif isclass(annotation) and issubclass(annotation, Event):
            return (annotation,)

        else:
            return ()

    async def resolve_dependency(
        event: Event, dependency: Dependency, stack: AsyncExitStack
    ) -> Any:
        func = dependency.dependency

        kwargs: dict[str, Any] = {}
        tasks: dict[str, asyncio.Task] = {}

        async with asyncio.TaskGroup() as tg:
            for param_name, param in signature(func).parameters.items():
                if isinstance(param.default, Dependency):
                    tasks[param_name] = tg.create_task(
                        resolve_dependency(
                            event=event, dependency=param.default, stack=stack
                        )
                    )

                elif isinstance(event, annotation_event_type(param.annotation)):
                    kwargs[param_name] = event

                elif param.default is not Parameter.empty:
                    kwargs[param_name] = param.default

                elif param.kind in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD):
                    pass

                else:
                    raise ValueError(
                        f"cannot resolve dependency for parameter '{param_name}' "
                        f"in function '{func.__name__}'. "
                        f"parameter must have either a default value, be an Event, or be a Dependency"
                    )

        kwargs.update({k: v.result() for k, v in tasks.items()})

        if isasyncgenfunction(func):
            return await stack.enter_async_context(asynccontextmanager(func)(**kwargs))

        elif isgeneratorfunction(func):
            return stack.enter_context(contextmanager(func)(**kwargs))

        elif iscoroutinefunction(func):
            return await func(**kwargs)

        else:
            return func(**kwargs)

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        sign = signature(func)

        event_type = ()

        for param in sign.parameters.values():
            event_type += annotation_event_type(param.annotation)

        if matcher:
            if any(
                param
                for param in sign.parameters.values()
                if isinstance(param.default, Dependency)
            ):

                @wraps(func)
                async def wrapper(event: Event, **kwargs) -> Any:
                    if isinstance(event, event_type) and matcher(event):
                        func_kwargs: dict[str, Any] = {}
                        tasks: dict[str, asyncio.Task] = {}

                        async with AsyncExitStack() as stack:
                            async with asyncio.TaskGroup() as tg:
                                for param_name, param in sign.parameters.items():
                                    if isinstance(param.default, Dependency):
                                        tasks[param_name] = tg.create_task(
                                            resolve_dependency(
                                                event=event,
                                                dependency=param.default,
                                                stack=stack,
                                            )
                                        )

                                    elif isinstance(
                                        event, annotation_event_type(param.annotation)
                                    ):
                                        func_kwargs[param_name] = event

                                    elif param.default is not Parameter.empty:
                                        func_kwargs[param_name] = param.default

                                    elif param.kind in (
                                        Parameter.VAR_POSITIONAL,
                                        Parameter.VAR_KEYWORD,
                                    ):
                                        pass

                                    else:
                                        raise ValueError(
                                            f"cannot resolve dependency for parameter '{param_name}' "
                                            f"in function '{func.__name__}'. "
                                            f"parameter must have either a default value, be an Event, or be a Dependency."
                                        )

                            return await func(
                                **{k: v.result() for k, v in tasks.items()},
                                **func_kwargs,
                                **kwargs,
                            )

            else:

                @wraps(func)
                async def wrapper(event: Event, **kwargs) -> Any:
                    if isinstance(event, event_type) and matcher(event):
                        return await func(event, **kwargs)

        else:
            if any(
                param
                for param in sign.parameters.values()
                if isinstance(param.default, Dependency)
            ):

                @wraps(func)
                async def wrapper(event: Event, **kwargs) -> Any:
                    if isinstance(event, event_type):
                        func_kwargs: dict[str, Any] = {}
                        tasks: dict[str, asyncio.Task] = {}

                        async with AsyncExitStack() as stack:
                            async with asyncio.TaskGroup() as tg:
                                for param_name, param in sign.parameters.items():
                                    if isinstance(param.default, Dependency):
                                        tasks[param_name] = tg.create_task(
                                            resolve_dependency(
                                                event=event,
                                                dependency=param.default,
                                                stack=stack,
                                            )
                                        )

                                    elif isinstance(
                                        event, annotation_event_type(param.annotation)
                                    ):
                                        func_kwargs[param_name] = event

                                    elif param.default is not Parameter.empty:
                                        func_kwargs[param_name] = param.default

                                    elif param.kind in (
                                        Parameter.VAR_POSITIONAL,
                                        Parameter.VAR_KEYWORD,
                                    ):
                                        pass

                                    else:
                                        raise ValueError(
                                            f"cannot resolve dependency for parameter '{param_name}' "
                                            f"in function '{func.__name__}'. "
                                            f"parameter must have either a default value, be an Event, or be a Dependency."
                                        )

                            return await func(
                                **{k: v.result() for k, v in tasks.items()},
                                **func_kwargs,
                                **kwargs,
                            )

            else:

                @wraps(func)
                async def wrapper(event: Event, **kwargs) -> Any:
                    if isinstance(event, event_type):
                        return await func(event, **kwargs)

        PluginManager.plugins[func.__module__].executors.append(wrapper)

        return wrapper

    return decorator



from dataclasses import KW_ONLY, dataclass, field
from typing import Any, ClassVar, Literal, TypeAlias

Context: TypeAlias = dict[str, Any]


@dataclass
class Event:
    _: KW_ONLY

    ctx: Context = field(repr=False)

    post_type: Literal["message", "notice", "request", "meta_event"]

    time: int
    self_id: int

    event_type: ClassVar[dict[str, type["Event"]]] = {}

    def __init_subclass__(cls, *args, **kwargs) -> None:
        Event.event_type[cls.post_type] = cls

    @classmethod
    def from_ctx(cls, *, ctx: Context) -> "Event":
        return (
            event.from_ctx(ctx=ctx)
            if (event := cls.event_type.get(ctx["post_type"]))
            else cls(
                ctx=ctx,
                time=ctx["time"],
                self_id=ctx["self_id"],
                post_type=ctx["post_type"],
            )
        )



import asyncio
import logging
from dataclasses import KW_ONLY, asdict, dataclass
from functools import cached_property
from typing import Any, ClassVar, Iterable, Literal, Self, override

from fastbot.bot import FastBot
from fastbot.event import Context, Event
from fastbot.message import Message, MessageSegment


@dataclass
class MessageEvent(Event):
    _: KW_ONLY

    post_type: Literal["message"] = "message"

    message_type: Literal["group", "private"]

    event_type: ClassVar[dict[str, type["MessageEvent"]]] = {}

    def __init_subclass__(cls, *args, **kwargs) -> None:
        MessageEvent.event_type[cls.message_type] = cls

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> "MessageEvent":
        return (
            event.from_ctx(ctx=ctx)
            if (event := cls.event_type.get(ctx["message_type"]))
            else cls(
                ctx=ctx,
                time=ctx["time"],
                self_id=ctx["self_id"],
                post_type=ctx["post_type"],
                message_type=ctx["message_type"],
            )
        )


@dataclass
class PrivateMessageEvent(MessageEvent):
    @dataclass
    class Sender:
        _: KW_ONLY

        user_id: int | None = None
        nickname: str | None = None
        sex: str | None = None
        age: int | None = None

    _: KW_ONLY

    message_type: Literal["private"] = "private"

    sub_type: Literal["friend", "group", "other"]

    message_id: int
    user_id: int
    message: Message
    raw_message: str
    font: int
    sender: Sender

    futures: ClassVar[dict[int, asyncio.Future]] = {}

    def __post_init__(self) -> None:
        logging.debug(self.__repr__())

        self.message = Message(
            MessageSegment(type=msg["type"], data=msg["data"]) for msg in self.message
        )

        self.sender = self.Sender(
            **{
                k: v
                for k, v in self.ctx["sender"].items()
                if k in self.Sender.__dataclass_fields__
            }
        )

        if future := self.__class__.futures.get(self.user_id):
            future.set_result(self)

    def __hash__(self) -> int:
        return hash((self.user_id, self.time, self.self_id, self.raw_message))

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> Self:
        return cls(
            ctx=ctx, **{k: v for k, v in ctx.items() if k in cls.__dataclass_fields__}
        )

    async def send(
        self,
        message: str
        | Message
        | MessageSegment
        | Iterable[str | Message | MessageSegment],
    ) -> Any:
        return await FastBot.do(
            endpoint="send_private_msg",
            message=[asdict(msg) for msg in Message(message)],
            self_id=self.self_id,
            user_id=self.user_id,
        )

    async def defer(
        self,
        message: str
        | Message
        | MessageSegment
        | Iterable[str | Message | MessageSegment],
    ) -> Self:
        self.__class__.futures[self.user_id] = future = asyncio.Future()

        await self.send(message=message)

        try:
            return await future
        finally:
            del self.__class__.futures[self.user_id]

    @cached_property
    def plaintext(self) -> str:
        return "".join(
            segment.data["text"] for segment in self.message if segment.type == "text"
        )


@dataclass
class GroupMessageEvent(MessageEvent):
    @dataclass
    class Sender:
        _: KW_ONLY

        user_id: int | None = None
        nickname: str | None = None
        card: str | None = None
        role: str | None = None
        sex: str | None = None
        age: int | None = None
        area: str | None = None
        level: str | None = None
        title: str | None = None

    _: KW_ONLY

    message_type: Literal["group"] = "group"

    sub_type: Literal["normal", "anonymous", "notice"]

    message_id: int
    group_id: int
    user_id: int
    message: Message
    raw_message: str
    font: int
    sender: Sender

    futures: ClassVar[dict[tuple[int, int], asyncio.Future]] = {}

    def __post_init__(self) -> None:
        logging.debug(self.__repr__())

        self.message = Message(
            MessageSegment(type=msg["type"], data=msg["data"]) for msg in self.message
        )

        self.sender = self.Sender(
            **{
                k: v
                for k, v in self.ctx["sender"].items()
                if k in self.Sender.__dataclass_fields__
            }
        )

        if future := self.__class__.futures.get((self.group_id, self.user_id)):
            future.set_result(self)

    def __hash__(self) -> int:
        return hash((self.user_id, self.time, self.self_id, self.raw_message))

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> Self:
        return cls(
            ctx=ctx, **{k: v for k, v in ctx.items() if k in cls.__dataclass_fields__}
        )

    async def send(
        self,
        message: str
        | Message
        | MessageSegment
        | Iterable[str | Message | MessageSegment],
    ) -> Any:
        return await FastBot.do(
            endpoint="send_group_msg",
            message=[asdict(msg) for msg in Message(message)],
            self_id=self.self_id,
            group_id=self.group_id,
        )

    async def defer(
        self,
        message: str
        | Message
        | MessageSegment
        | Iterable[str | Message | MessageSegment],
    ) -> Self:
        self.__class__.futures[(self.group_id, self.user_id)] = future = (
            asyncio.Future()
        )

        await self.send(message=message)

        try:
            return await future
        finally:
            del self.__class__.futures[(self.group_id, self.user_id)]

    @cached_property
    def plaintext(self) -> str:
        return "".join(
            segment.data["text"] for segment in self.message if segment.type == "text"
        )



import logging
from dataclasses import KW_ONLY, dataclass
from typing import ClassVar, Literal, Self, override

from fastbot.event import Context, Event


@dataclass
class NoticeEvent(Event):
    _: KW_ONLY

    post_type: Literal["notice"] = "notice"

    notice_type: str

    event_type: ClassVar[dict[str, type["NoticeEvent"]]] = {}

    def __init_subclass__(cls, *args, **kwargs) -> None:
        NoticeEvent.event_type[cls.notice_type] = cls

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> "NoticeEvent":
        return (
            event.from_ctx(ctx=ctx)
            if (event := cls.event_type.get(ctx["notice_type"]))
            else cls(
                ctx=ctx,
                time=ctx["time"],
                self_id=ctx["self_id"],
                post_type=ctx["post_type"],
                notice_type=ctx["notice_type"],
            )
        )


@dataclass
class GroupFileUploadNoticeEvent(NoticeEvent):
    @dataclass
    class File:
        _: KW_ONLY

        id: str
        name: str
        size: int
        busid: int

    _: KW_ONLY

    notice_type: Literal["group_upload"] = "group_upload"

    group_id: int
    user_id: int
    file: File

    def __post_init__(self) -> None:
        logging.debug(self.__repr__())

        self.file = self.File(
            **{
                k: v
                for k, v in self.ctx["file"].items()
                if k in self.File.__dataclass_fields__
            }
        )

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> Self:
        return cls(
            ctx=ctx, **{k: v for k, v in ctx.items() if k in cls.__dataclass_fields__}
        )


@dataclass
class GroupAdminChangeNoticeEvent(NoticeEvent):
    _: KW_ONLY

    notice_type: Literal["group_admin"] = "group_admin"

    sub_type: Literal["set", "unset"]

    group_id: int
    user_id: int

    def __post_init__(self) -> None:
        logging.debug(self.__repr__())

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> Self:
        return cls(
            ctx=ctx, **{k: v for k, v in ctx.items() if k in cls.__dataclass_fields__}
        )


@dataclass
class GroupMemberDecreaseNoticeEvent(NoticeEvent):
    _: KW_ONLY

    notice_type: Literal["group_decrease"] = "group_decrease"

    sub_type: Literal["leave", "kick", "kick_me"]

    group_id: int
    user_id: int
    operator_id: int

    def __post_init__(self) -> None:
        logging.debug(self.__repr__())

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> Self:
        return cls(
            ctx=ctx, **{k: v for k, v in ctx.items() if k in cls.__dataclass_fields__}
        )


@dataclass
class GroupMemberIncreaseNoticeEvent(NoticeEvent):
    _: KW_ONLY

    notice_type: Literal["group_increase"] = "group_increase"

    sub_type: Literal["approve", "invite"]

    group_id: int
    operator_id: int
    user_id: int

    def __post_init__(self) -> None:
        logging.debug(self.__repr__())

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> Self:
        return cls(
            ctx=ctx, **{k: v for k, v in ctx.items() if k in cls.__dataclass_fields__}
        )


@dataclass
class GroupBanNoticeEvent(NoticeEvent):
    _: KW_ONLY

    notice_type: Literal["group_ban"] = "group_ban"

    sub_type: Literal["ban", "lift_ban"]

    group_id: int
    operator_id: int
    user_id: int
    duration: int

    def __post_init__(self) -> None:
        logging.debug(self.__repr__())

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> Self:
        return cls(
            ctx=ctx, **{k: v for k, v in ctx.items() if k in cls.__dataclass_fields__}
        )


@dataclass
class FriendAddNoticeEvent(NoticeEvent):
    _: KW_ONLY

    notice_type: Literal["friend_add"] = "friend_add"

    user_id: int

    def __post_init__(self) -> None:
        logging.debug(self.__repr__())

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> Self:
        return cls(
            ctx=ctx, **{k: v for k, v in ctx.items() if k in cls.__dataclass_fields__}
        )


@dataclass
class GroupMessageRecallNoticeEvent(NoticeEvent):
    _: KW_ONLY

    notice_type: Literal["group_recall"] = "group_recall"

    group_id: int
    user_id: int
    operator_id: int
    message_id: int

    def __post_init__(self) -> None:
        logging.debug(self.__repr__())

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> Self:
        return cls(
            ctx=ctx, **{k: v for k, v in ctx.items() if k in cls.__dataclass_fields__}
        )


@dataclass
class FriendMessageRecallNoticeEvent(NoticeEvent):
    _: KW_ONLY

    notice_type: Literal["friend_recall"] = "friend_recall"

    user_id: int
    message_id: int

    def __post_init__(self) -> None:
        logging.debug(self.__repr__())

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> Self:
        return cls(
            ctx=ctx, **{k: v for k, v in ctx.items() if k in cls.__dataclass_fields__}
        )



import logging
from dataclasses import KW_ONLY, dataclass
from typing import Any, ClassVar, Literal, Self, override

from fastbot.bot import FastBot
from fastbot.event import Context, Event


@dataclass
class RequestEvent(Event):
    _: KW_ONLY

    post_type: Literal["request"] = "request"

    request_type: Literal["friend", "group"]

    event_type: ClassVar[dict[str, type["RequestEvent"]]] = {}

    def __init_subclass__(cls, *args, **kwargs) -> None:
        RequestEvent.event_type[cls.request_type] = cls

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> "RequestEvent":
        return (
            event.from_ctx(ctx=ctx)
            if (event := cls.event_type.get(ctx["request_type"]))
            else cls(
                ctx=ctx,
                time=ctx["time"],
                self_id=ctx["self_id"],
                post_type=ctx["post_type"],
                request_type=ctx["request_type"],
            )
        )


@dataclass
class FriendRequestEvent(RequestEvent):
    _: KW_ONLY

    request_type: Literal["friend"] = "friend"

    user_id: int
    comment: str
    flag: str

    def __post_init__(self) -> None:
        logging.debug(self.__repr__())

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> Self:
        return cls(
            ctx=ctx, **{k: v for k, v in ctx.items() if k in cls.__dataclass_fields__}
        )

    async def approve(self, *, remark: str | None = None) -> Any:
        return await FastBot.do(
            endpoint="set_friend_add_request",
            self_id=self.self_id,
            approve=True,
            flag=self.flag,
            remark=remark,
        )

    async def reject(self) -> Any:
        return await FastBot.do(
            endpoint="set_friend_add_request", self_id=self.self_id, approve=False
        )


@dataclass
class GroupRequestEvent(RequestEvent):
    _: KW_ONLY

    request_type: Literal["group"] = "group"

    sub_type: Literal["add", "invite"]

    group_id: int
    user_id: int
    comment: str
    flag: str

    def __post_init__(self) -> None:
        logging.debug(self.__repr__())

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> Self:
        return cls(
            ctx=ctx, **{k: v for k, v in ctx.items() if k in cls.__dataclass_fields__}
        )

    async def approve(self) -> Any:
        return await FastBot.do(
            endpoint="set_group_add_request",
            self_id=self.self_id,
            approve=True,
            sub_type=self.sub_type,
        )

    async def reject(self, *, reason: str | None = None) -> Any:
        return await FastBot.do(
            endpoint="set_group_add_request",
            self_id=self.self_id,
            approve=False,
            reason=reason,
        )



import logging
from dataclasses import KW_ONLY, dataclass
from typing import ClassVar, Literal, Self, override

from fastbot.event import Context, Event


@dataclass
class MetaEvent(Event):
    _: KW_ONLY

    post_type: Literal["meta_event"] = "meta_event"

    meta_event_type: Literal["heartbeat", "lifecycle"]

    event_type: ClassVar[dict[str, type["MetaEvent"]]] = {}

    def __init_subclass__(cls, *args, **kwargs) -> None:
        MetaEvent.event_type[cls.meta_event_type] = cls

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> "MetaEvent":
        return (
            event.from_ctx(ctx=ctx)
            if (event := cls.event_type.get(ctx["meta_event_type"]))
            else cls(
                ctx=ctx,
                time=ctx["time"],
                self_id=ctx["self_id"],
                post_type=ctx["post_type"],
                meta_event_type=ctx["meta_event_type"],
            )
        )


@dataclass
class LifecycleMetaEvent(MetaEvent):
    _: KW_ONLY

    meta_event_type: Literal["lifecycle"] = "lifecycle"

    sub_type: Literal["enable", "disable", "connect"]

    def __post_init__(self) -> None:
        logging.debug(self.__repr__())

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> Self:
        return cls(
            ctx=ctx, **{k: v for k, v in ctx.items() if k in cls.__dataclass_fields__}
        )


@dataclass
class HeartbeatMetaEvent(MetaEvent):
    _: KW_ONLY

    meta_event_type: Literal["heartbeat"] = "heartbeat"

    status: dict
    interval: int

    def __post_init__(self) -> None:
        logging.debug(self.__repr__())

    @classmethod
    @override
    def from_ctx(cls, *, ctx: Context) -> Self:
        return cls(
            ctx=ctx, **{k: v for k, v in ctx.items() if k in cls.__dataclass_fields__}
        )



import asyncio
from contextlib import AsyncExitStack, asynccontextmanager
from inspect import isasyncgenfunction
import logging
import os
from contextvars import ContextVar
from functools import partial
from typing import Any, AsyncGenerator, ClassVar, Iterable, Self
from weakref import WeakValueDictionary

import uvicorn
from fastapi import FastAPI, WebSocket, WebSocketException, status

from fastbot.plugin import PluginManager

try:
    import ujson as json

    json.dumps = partial(json.dumps, ensure_ascii=False, sort_keys=False)

except ImportError:
    import json

    json.dumps = partial(
        json.dumps, ensure_ascii=False, separators=(",", ":"), sort_keys=False
    )


class FastBot:
    __slots__ = ()

    app: ClassVar[FastAPI]

    connectors: ClassVar[WeakValueDictionary[int, WebSocket]] = WeakValueDictionary()
    futures: ClassVar[dict[int, asyncio.Future]] = {}

    self_id: ClassVar[ContextVar[int | None]] = ContextVar("self_id", default=None)

    @classmethod
    async def ws_adapter(cls, websocket: WebSocket) -> None:
        if authorization := os.getenv("FASTBOT_AUTHORIZATION"):
            if not (access_token := websocket.headers.get("authorization")):
                raise WebSocketException(
                    code=status.WS_1008_POLICY_VIOLATION,
                    reason="missing `authorization` header",
                )

            match access_token.split():
                case [header, token] if header.title() in ("Bearer", "Token"):
                    if token != authorization:
                        raise WebSocketException(
                            code=status.HTTP_403_FORBIDDEN,
                            reason="invalid `authorization` header",
                        )

                case [token]:
                    if token != authorization:
                        raise WebSocketException(
                            code=status.HTTP_403_FORBIDDEN,
                            reason="invalid `authorization` header",
                        )

                case _:
                    raise WebSocketException(
                        code=status.HTTP_403_FORBIDDEN,
                        reason="invalid `authorization` header",
                    )

        if not (self_id := websocket.headers.get("x-self-id")):
            raise WebSocketException(
                code=status.WS_1008_POLICY_VIOLATION,
                reason="missing `x-self-id` header",
            )

        if not (self_id.isdigit() and (self_id := int(self_id))):
            raise WebSocketException(
                code=status.WS_1008_POLICY_VIOLATION,
                reason="invalid `x-self-id` header",
            )

        if self_id in cls.connectors:
            raise WebSocketException(
                code=status.WS_1008_POLICY_VIOLATION,
                reason="duplicate `x-self-id` header",
            )

        await websocket.accept()

        logging.info(f"websocket connected {self_id=}")

        cls.connectors[self_id] = websocket

        await cls.event_handler(websocket=websocket)

    @classmethod
    async def event_handler(cls, websocket: WebSocket) -> None:
        async with asyncio.TaskGroup() as tg:
            while True:
                match message := await websocket.receive():
                    case {"bytes": data} | {"text": data}:
                        if "post_type" in (ctx := json.loads(data)):
                            cls.self_id.set(ctx.get("self_id"))

                            tg.create_task(PluginManager.run(ctx=ctx))

                        elif ctx["status"] == "ok":
                            cls.futures[ctx["echo"]].set_result(ctx.get("data"))

                        else:
                            cls.futures[ctx["echo"]].set_exception(RuntimeError(ctx))

                    case _:
                        logging.warning(f"unknow websocket message received {message=}")

    @classmethod
    async def do(cls, *, endpoint: str, self_id: int | None = None, **kwargs) -> Any:
        if not (
            self_id := (
                self_id
                or cls.self_id.get()
                or (next(iter(cls.connectors)) if len(cls.connectors) == 1 else None)
            )
        ):
            raise RuntimeError("parameter `self_id` must be specified")

        logging.debug(f"{endpoint=} {self_id=} {kwargs=}")

        future = asyncio.Future()
        future_id = id(future)

        cls.futures[future_id] = future

        try:
            await cls.connectors[self_id].send_bytes(
                json.dumps(
                    {"action": endpoint, "params": kwargs, "echo": future_id}
                ).encode(encoding="utf-8")
            )

            return await future

        finally:
            del cls.futures[future_id]

    @classmethod
    def build(cls, plugins: str | Iterable[str] | None = None, **kwargs) -> Self:
        if isinstance(plugins, str):
            PluginManager.import_from(plugins)

        elif isinstance(plugins, Iterable):
            for plugin in plugins:
                PluginManager.import_from(plugin)

        @asynccontextmanager
        async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
            app.add_api_websocket_route("/onebot/v11/ws", cls.ws_adapter)

            async with AsyncExitStack() as stack, asyncio.TaskGroup() as tg:
                if lifespan := kwargs.pop("lifespan", None):
                    await stack.enter_async_context(lifespan(app))

                await asyncio.gather(
                    *(
                        (
                            stack.enter_async_context(asynccontextmanager(init)())
                            if isasyncgenfunction(init)
                            else init()
                        )
                        for plugin in PluginManager.plugins.values()
                        if (init := plugin.init)
                    )
                )

                for plugin in PluginManager.plugins.values():
                    if background := plugin.backgrounds:
                        for task in background:
                            tg.create_task(task())

                yield

        app = FastAPI(lifespan=lifespan, **kwargs)

        cls.app = app

        return cls()

    @classmethod
    def run(cls, **kwargs) -> None:
        uvicorn.run(app=cls.app, **kwargs)



from base64 import b64encode
from dataclasses import KW_ONLY, asdict, dataclass
from itertools import chain, groupby
from operator import attrgetter
from typing import Any, Iterable, Iterator, Literal, Self, TypedDict, Union

PREV, NEXT, DATA = 0, 1, 2


class MessageSegmentData(TypedDict):
    type: str
    data: dict[str, Any]


class Link(list):
    def __init__(
        self, iterable: Iterable[Any] | None = None, maxlen: int | None = None
    ) -> None:
        self[:] = [self, self, None]

        self.length = 0
        self.maxlen = maxlen

        if iterable:
            self.extend(iterable)

    def __iter__(self) -> Iterator[Any]:
        next_node = self[NEXT]

        while next_node is not self:
            yield next_node[DATA]
            next_node = next_node[NEXT]

    def __len__(self) -> int:
        return self.length

    def __delitem__(self, idx: int) -> None:
        if not (-self.length <= idx < self.length):
            raise IndexError("index out of range")

        if idx < 0:
            idx += self.length

        if idx <= self.length // 2:
            current_node = self[NEXT]
            for _ in range(idx):
                current_node = current_node[NEXT]
        else:
            current_node = self[PREV]
            for _ in range(self.length - idx - 1):
                current_node = current_node[PREV]

        prev_node = current_node[PREV]
        next_node = current_node[NEXT]
        prev_node[NEXT] = next_node
        next_node[PREV] = prev_node

        self.length -= 1

    def append(self, value: Any) -> None:
        if self.maxlen is not None and self.length == self.maxlen:
            self.popleft()

        prev_node = self[PREV]
        prev_node[NEXT] = self[PREV] = [prev_node, self, value]

        self.length += 1

    def appendleft(self, value: Any) -> None:
        if self.maxlen is not None and self.length == self.maxlen:
            self.pop()

        next_node = self[NEXT]
        next_node[PREV] = self[NEXT] = [self, next_node, value]

        self.length += 1

    def clear(self) -> None:
        self[:] = [self, self, None]
        self.length = 0

    def extend(self, values: Any) -> None:
        for value in values:
            self.append(value)

    def extendleft(self, values: Any) -> None:
        for value in values:
            self.appendleft(value)

    def insert(self, idx: int, value: Any) -> None:
        if not (-self.length <= idx < self.length):
            raise IndexError("index out of range")

        if self.maxlen is not None and self.length == self.maxlen:
            raise IndexError("list has reached its maximum length")

        if idx == 0:
            self.appendleft(value)

        elif idx < 0:
            idx += self.length

        if idx <= self.length // 2:
            current_node = self[NEXT]
            for _ in range(idx):
                current_node = current_node[NEXT]

        else:
            current_node = self
            for _ in range(self.length - idx):
                current_node = current_node[PREV]

        prev_node = current_node
        next_node = current_node[NEXT]
        link = [prev_node, next_node, value]
        next_node[PREV] = link
        prev_node[NEXT] = link

        self.length += 1

    def pop(self) -> Any:
        if self.length == 0:
            raise IndexError("pop from empty list")

        prev_node = self[PREV]
        prev_node[PREV][NEXT] = self
        self[PREV] = prev_node[PREV]

        self.length -= 1
        return prev_node[DATA]

    def popleft(self) -> Any:
        if self.length == 0:
            raise IndexError("pop from empty list")

        next_node = self[NEXT]
        next_node[NEXT][PREV] = self
        self[NEXT] = next_node[NEXT]

        self.length -= 1
        return next_node[DATA]

    def reverse(self) -> None:
        prev_node = self
        current_node = self[NEXT]
        self[NEXT] = self[PREV] = self

        while current_node is not self:
            next_node = current_node[NEXT]
            current_node[NEXT] = prev_node
            current_node[PREV] = prev_node
            prev_node = current_node
            current_node = next_node

        self[NEXT] = prev_node
        self[PREV] = prev_node[PREV]

    def remove(self, value: Any) -> None:
        current_node = self[NEXT]
        while current_node is not self:
            if current_node[DATA] == value:
                prev_node = current_node[PREV]
                next_node = current_node[NEXT]
                prev_node[NEXT] = next_node
                next_node[PREV] = prev_node

                self.length -= 1
                return

            current_node = current_node[NEXT]

        raise ValueError(f"{value} is not in list")

    def rotate(self, n: int) -> None:
        if (length := self.length) == 0:
            return

        n %= length
        if n == 0:
            return

        if n < 0:
            n += length

        if n <= length // 2:
            for _ in range(n):
                prev_node = self[PREV]
                self[PREV] = prev_node[PREV]
                prev_node[PREV][NEXT] = self
                self[NEXT][PREV] = prev_node
                prev_node[NEXT] = self[NEXT]
                self[NEXT] = prev_node
                prev_node[PREV] = self

        else:
            for _ in range(length - n):
                next_node = self[NEXT]
                self[NEXT] = next_node[NEXT]
                next_node[NEXT][PREV] = self
                self[PREV][NEXT] = next_node
                next_node[PREV] = self[PREV]
                self[PREV] = next_node
                next_node[NEXT] = self


@dataclass(slots=True)
class MessageSegment:
    _: KW_ONLY

    type: str
    data: dict[str, Any]

    def __add__(self, other: Union[str, Iterable[Any], "MessageSegment"]) -> "Message":
        return Message(content=self) + other

    def __radd__(self, other: Union[str, Iterable[Any], "MessageSegment"]) -> "Message":
        return Message(content=other) + self

    @classmethod
    def text(cls, text: str) -> Self:
        return cls(type="text", data={"text": text})

    @classmethod
    def face(cls, id: int) -> Self:
        return cls(type="face", data={"id": str(id)})

    @classmethod
    def image(
        cls,
        file: bytes | str,
        type: Literal["flash"] | None = None,
        url: str | None = None,
        cache: bool | None = None,
        proxy: bool | None = None,
        timeout: int | None = None,
    ) -> Self:
        return cls(
            type="image",
            data={
                "file": (
                    file
                    if isinstance(file, str)
                    else f"base64://{b64encode(file).decode(encoding='utf-8')}"
                ),
                "type": type,
                "url": url,
                "cache": cache,
                "proxy": proxy,
                "timeout": timeout,
            },
        )

    @classmethod
    def record(
        cls,
        file: str,
        magic: bool | None = None,
        url: str | None = None,
        cache: bool | None = None,
        proxy: bool | None = None,
        timeout: int | None = None,
    ) -> Self:
        return cls(
            type="record",
            data={
                "file": file,
                "magic": magic,
                "url": url,
                "cache": cache,
                "proxy": proxy,
                "timeout": timeout,
            },
        )

    @classmethod
    def video(
        cls,
        file: str,
        url: str | None = None,
        cache: bool | None = None,
        proxy: bool | None = None,
        timeout: int | None = None,
    ) -> Self:
        return cls(
            type="video",
            data={
                "file": file,
                "url": url,
                "cache": cache,
                "proxy": proxy,
                "timeout": timeout,
            },
        )

    @classmethod
    def at(cls, qq: int | Literal["all"]) -> Self:
        return cls(type="at", data={"qq": str(qq)})

    @classmethod
    def reply(cls, id: int) -> Self:
        return cls(type="reply", data={"id": str(id)})

    @classmethod
    def forward(cls, id: int) -> Self:
        return cls(type="forward", data={"id": str(id)})

    @classmethod
    def node(
        cls,
        id: int | None = None,
        content: Union["Message", list[Union["MessageSegment", MessageSegmentData]]]
        | None = None,
        **kwargs,
    ) -> Self:
        if id:
            return cls(type="node", data={"id": str(id)})
        elif content:
            return cls(
                type="node",
                data={
                    "content": [
                        asdict(segment)
                        if isinstance(segment, MessageSegment)
                        else segment
                        for segment in content
                    ],
                    **kwargs,
                },
            )
        else:
            raise ValueError("parameter `id` or `content` must be specified")


class Message(Link):
    def __init__(
        self, content: str | Iterable[Any] | MessageSegment | None = None
    ) -> None:
        super().__init__()

        if content:
            if isinstance(content, MessageSegment):
                self.append(content)
            elif isinstance(content, str):
                self.append(MessageSegment.text(text=content))
            elif isinstance(content, Iterable):
                self.extend(
                    chain.from_iterable(Message(content=item) for item in content)
                )
            else:
                raise ValueError("unsupported message type")

    def __add__(self, other: str | Iterable[Any] | MessageSegment) -> "Message":
        message = Message(content=self)
        message += other

        return message

    def __radd__(self, other: str | Iterable[Any] | MessageSegment) -> "Message":
        message = Message(content=other)
        message += self

        return message

    def __iadd__(self, other: str | Iterable[Any] | MessageSegment) -> "Message":
        if isinstance(other, Message):
            self.extend(other)
        elif isinstance(other, MessageSegment):
            self.append(other)
        elif isinstance(other, str):
            self.append(MessageSegment.text(text=other))
        elif isinstance(other, Iterable):
            self.extend(chain.from_iterable(Message(content=item) for item in other))
        else:
            raise ValueError("unsupported message type")

        return self

    def compact(self, *, concat: str = "") -> "Message":
        return Message(
            MessageSegment.text(
                text=concat.join(segment.data["text"] for segment in segments)
            )
            if key == "text"
            else segments
            for key, segments in groupby(self, key=attrgetter("type"))
        )



from dataclasses import KW_ONLY, dataclass, field
from typing import Callable


@dataclass
class Matcher:
    rule: Callable[..., bool] = field(default=lambda: True)

    _: KW_ONLY

    matchers: list["Matcher"] = field(default_factory=list)
    operator: str | None = field(default=None)

    def __call__(self, *args, **kwargs) -> bool:
        match self.operator:
            case "and":
                return all(matcher(*args, **kwargs) for matcher in self.matchers)

            case "or":
                return any(matcher(*args, **kwargs) for matcher in self.matchers)

            case _:
                return self.rule(*args, **kwargs)

    def __and__(self, other: "Matcher") -> "Matcher":
        if self.operator == "and":
            self.matchers.append(other)

            return self

        else:
            return Matcher(matchers=[self, other], operator="and")

    def __or__(self, other: "Matcher") -> "Matcher":
        if self.operator == "or":
            self.matchers.append(other)

            return self

        else:
            return Matcher(matchers=[self, other], operator="or")

    def __invert__(self) -> "Matcher":
        return Matcher(rule=lambda *args, **kwargs: not self(*args, **kwargs))
